@@ -34,6 +34,7 @@
 #include <libgen.h>
 #include <signal.h>
 #include <limits.h>
+#include <ctype.h>
 
 #include <ftw.h>
 
@@ -103,58 +104,70 @@ enum pwrap_dbglvl_e {
 	PWRAP_LOG_TRACE
 };
 
-#ifdef NDEBUG
-# define PWRAP_LOG(...)
-#else /* NDEBUG */
 static void pwrap_log(enum pwrap_dbglvl_e dbglvl,
 		      const char *function,
 		      const char *format, ...) PRINTF_ATTRIBUTE(3, 4);
 # define PWRAP_LOG(dbglvl, ...) pwrap_log((dbglvl), __func__, __VA_ARGS__)
 
-static void pwrap_log(enum pwrap_dbglvl_e dbglvl,
-		      const char *function,
-		      const char *format, ...)
+static void pwrap_vlog(enum pwrap_dbglvl_e dbglvl,
+		       const char *function,
+		       const char *format,
+		       va_list args) PRINTF_ATTRIBUTE(3, 0);
+
+static void pwrap_vlog(enum pwrap_dbglvl_e dbglvl,
+		       const char *function,
+		       const char *format,
+		       va_list args)
 {
 	char buffer[1024];
-	va_list va;
 	const char *d;
 	unsigned int lvl = 0;
+	const char *prefix = "PWRAP";
 
 	d = getenv("PAM_WRAPPER_DEBUGLEVEL");
 	if (d != NULL) {
 		lvl = atoi(d);
 	}
 
-	va_start(va, format);
-	vsnprintf(buffer, sizeof(buffer), format, va);
-	va_end(va);
-
-	if (lvl >= dbglvl) {
-		const char *prefix = "PWRAP";
-		switch (dbglvl) {
-			case PWRAP_LOG_ERROR:
-				prefix = "PWRAP_ERROR";
-				break;
-			case PWRAP_LOG_WARN:
-				prefix = "PWRAP_WARN";
-				break;
-			case PWRAP_LOG_DEBUG:
-				prefix = "PWRAP_DEBUG";
-				break;
-			case PWRAP_LOG_TRACE:
-				prefix = "PWRAP_TRACE";
-				break;
-		}
-
-		fprintf(stderr,
-			"%s(%d) - %s: %s\n",
-			prefix,
-			(int)getpid(),
-			function,
-			buffer);
+	if (lvl < dbglvl) {
+		return;
 	}
+
+	vsnprintf(buffer, sizeof(buffer), format, args);
+
+	switch (dbglvl) {
+	case PWRAP_LOG_ERROR:
+		prefix = "PWRAP_ERROR";
+		break;
+	case PWRAP_LOG_WARN:
+		prefix = "PWRAP_WARN";
+		break;
+	case PWRAP_LOG_DEBUG:
+		prefix = "PWRAP_DEBUG";
+		break;
+	case PWRAP_LOG_TRACE:
+		prefix = "PWRAP_TRACE";
+		break;
+	}
+
+	fprintf(stderr,
+		"%s(%d) - %s: %s\n",
+		prefix,
+		(int)getpid(),
+		function,
+		buffer);
+}
+
+static void pwrap_log(enum pwrap_dbglvl_e dbglvl,
+		      const char *function,
+		      const char *format, ...)
+{
+	va_list va;
+
+	va_start(va, format);
+	pwrap_vlog(dbglvl, function, format, va);
+	va_end(va);
 }
-#endif /* NDEBUG */
 
 /*****************
  * LIBC
@@ -509,7 +522,7 @@ static void libpam_pam_vsyslog(const pam_handle_t *pamh,
 						   fmt,
 						   args);
 }
-#endif
+#endif /* HAVE_PAM_VSYSLOG */
 
 /*********************************************************
  * PWRAP INIT
@@ -518,7 +531,7 @@ static void libpam_pam_vsyslog(const pam_handle_t *pamh,
 #define BUFFER_SIZE 32768
 
 /* copy file from src to dst, overwrites dst */
-static int p_copy(const char *src, const char *dst, const char *pdir, mode_t mode)
+static int p_copy(const char *src, const char *dst, mode_t mode)
 {
 	int srcfd = -1;
 	int dstfd = -1;
@@ -541,7 +554,8 @@ static int p_copy(const char *src, const char *dst, const char *pdir, mode_t mod
 	if (mode == 0) {
 		rc = fstat(srcfd, &sb);
 		if (rc != 0) {
-			return -1;
+			rc = -1;
+			goto out;
 		}
 		mode = sb.st_mode;
 	}
@@ -553,7 +567,6 @@ static int p_copy(const char *src, const char *dst, const char *pdir, mode_t mod
 	}
 
 	for (;;) {
-		char *p;
 		bread = read(srcfd, buf, BUFFER_SIZE);
 		if (bread == 0) {
 			/* done */
@@ -564,21 +577,6 @@ static int p_copy(const char *src, const char *dst, const char *pdir, mode_t mod
 			goto out;
 		}
 
-		/* EXTRA UGLY HACK */
-		if (pdir != NULL) {
-			p = buf;
-
-			while (p < buf + BUFFER_SIZE) {
-				if (*p == '/') {
-					cmp = memcmp(p, "/etc/pam.d", 10);
-					if (cmp == 0) {
-						memcpy(p, pdir, 10);
-					}
-				}
-				p++;
-			}
-		}
-
 		bwritten = write(dstfd, buf, bread);
 		if (bwritten < 0) {
 			errno = EIO;
@@ -654,7 +652,7 @@ static int copy_ftw(const char *fpath,
 	}
 
 	PWRAP_LOG(PWRAP_LOG_TRACE, "Copying %s", fpath);
-	rc = p_copy(fpath, buf, NULL, sb->st_mode);
+	rc = p_copy(fpath, buf, sb->st_mode);
 	if (rc != 0) {
 		return FTW_STOP;
 	}
@@ -742,18 +740,114 @@ static void pwrap_clean_stale_dirs(const char *dir)
 	return;
 }
 
+static int pso_copy(const char *src, const char *dst, const char *pdir, mode_t mode)
+{
+	int srcfd = -1;
+	int dstfd = -1;
+	int rc = -1;
+	ssize_t bread, bwritten;
+	struct stat sb;
+	char buf[10];
+	int cmp;
+	size_t to_read;
+	bool found_slash;
+
+	cmp = strcmp(src, dst);
+	if (cmp == 0) {
+		return -1;
+	}
+
+	srcfd = open(src, O_RDONLY, 0);
+	if (srcfd < 0) {
+		return -1;
+	}
+
+	if (mode == 0) {
+		rc = fstat(srcfd, &sb);
+		if (rc != 0) {
+			rc = -1;
+			goto out;
+		}
+		mode = sb.st_mode;
+	}
+
+	dstfd = open(dst, O_CREAT|O_WRONLY|O_TRUNC, mode);
+	if (dstfd < 0) {
+		rc = -1;
+		goto out;
+	}
+
+	found_slash = false;
+	to_read = 1;
+
+	for (;;) {
+		bread = read(srcfd, buf, to_read);
+		if (bread == 0) {
+			/* done */
+			break;
+		} else if (bread < 0) {
+			errno = EIO;
+			rc = -1;
+			goto out;
+		}
+
+		to_read = 1;
+		if (!found_slash && buf[0] == '/') {
+			found_slash = true;
+			to_read = 9;
+		}
+
+		if (found_slash && bread == 9) {
+			cmp = memcmp(buf, "etc/pam.d", 9);
+			if (cmp == 0) {
+				memcpy(buf, pdir + 1, 9);
+			}
+			found_slash = false;
+		}
+
+		bwritten = write(dstfd, buf, bread);
+		if (bwritten < 0) {
+			errno = EIO;
+			rc = -1;
+			goto out;
+		}
+
+		if (bread != bwritten) {
+			errno = EFAULT;
+			rc = -1;
+			goto out;
+		}
+	}
+
+	rc = 0;
+out:
+	if (srcfd != -1) {
+		close(srcfd);
+	}
+	if (dstfd != -1) {
+		close(dstfd);
+	}
+	if (rc < 0) {
+		unlink(dst);
+	}
+
+	return rc;
+}
+
 static void pwrap_init(void)
 {
 	char tmp_config_dir[] = "/tmp/pam.X";
 	size_t len = strlen(tmp_config_dir);
 	const char *env;
-	uint32_t i;
+	struct stat sb;
 	int rc;
+	unsigned i;
 	char pam_library[128] = { 0 };
 	char libpam_path[1024] = { 0 };
 	ssize_t ret;
 	FILE *pidfile;
 	char pidfile_path[1024] = { 0 };
+	char letter;
 
 	if (!pam_wrapper_enabled()) {
 		return;
@@ -763,33 +857,36 @@ static void pwrap_init(void)
 		return;
 	}
 
-	PWRAP_LOG(PWRAP_LOG_DEBUG, "Initialize pam_wrapper");
+	/*
+	 * The name is selected to match/replace /etc/pam.d
+	 * We start from a random alphanum trying letters until
+	 * an available directory is found.
+	 */
+	letter = 48 + (getpid() % 70);
+	for (i = 0; i < 127; i++) {
+		if (isalpha(letter) || isdigit(letter)) {
+			tmp_config_dir[len - 1] = letter;
 
-	for (i = 0; i < 36; i++) {
-		struct stat sb;
-		char c;
-
-		if (i < 10) {
-			c = (char)(i + 48);
-		} else {
-			c = (char)(i + 87);
+			rc = lstat(tmp_config_dir, &sb);
+			if (rc == 0) {
+				PWRAP_LOG(PWRAP_LOG_TRACE,
+					  "Check if pam_wrapper dir %s is a "
+					  "stale directory",
+					  tmp_config_dir);
+				pwrap_clean_stale_dirs(tmp_config_dir);
+			} else if (rc < 0) {
+				if (errno != ENOENT) {
+					continue;
+				}
+				break; /* found */
+			}
 		}
 
-		tmp_config_dir[len - 1] = c;
-		rc = lstat(tmp_config_dir, &sb);
-		if (rc == 0) {
-			PWRAP_LOG(PWRAP_LOG_TRACE,
-				  "Check if pam_wrapper dir %s is a "
-				  "stale directory",
-				  tmp_config_dir);
-			pwrap_clean_stale_dirs(tmp_config_dir);
-			continue;
-		} else if (errno == ENOENT) {
-			break;
-		}
+		letter++;
+		letter %= 127;
 	}
 
-	if (i == 36) {
+	if (i == 127) {
 		PWRAP_LOG(PWRAP_LOG_ERROR,
 			  "Failed to find a possible path to create "
 			  "pam_wrapper config dir: %s",
@@ -797,6 +894,10 @@ static void pwrap_init(void)
 		exit(1);
 	}
 
+	PWRAP_LOG(PWRAP_LOG_DEBUG, "Initialize pam_wrapper");
+
+	pwrap_clean_stale_dirs(tmp_config_dir);
+
 	pwrap.config_dir = strdup(tmp_config_dir);
 	if (pwrap.config_dir == NULL) {
 		PWRAP_LOG(PWRAP_LOG_ERROR,
@@ -885,11 +986,13 @@ static void pwrap_init(void)
 			 "%s",
 			 pam_library);
 	} else {
-		char libpam_path_cp[sizeof(libpam_path)];
-		char *dname;
+		char libpam_path_cp[1024] = {0};
+		char *dname = NULL;
 
-		strncpy(libpam_path_cp, libpam_path, sizeof(libpam_path_cp));
-		libpam_path_cp[sizeof(libpam_path_cp) - 1] = '\0';
+		snprintf(libpam_path_cp,
+			 sizeof(libpam_path_cp),
+			 "%s",
+			 libpam_path);
 
 		dname = dirname(libpam_path_cp);
 		if (dname == NULL) {
@@ -908,7 +1011,7 @@ static void pwrap_init(void)
 	PWRAP_LOG(PWRAP_LOG_TRACE, "Reconstructed PAM path: %s", libpam_path);
 
 	PWRAP_LOG(PWRAP_LOG_DEBUG, "Copy %s to %s", libpam_path, pwrap.libpam_so);
-	rc = p_copy(libpam_path, pwrap.libpam_so, pwrap.config_dir, 0644);
+	rc = pso_copy(libpam_path, pwrap.libpam_so, pwrap.config_dir, 0644);
 	if (rc != 0) {
 		PWRAP_LOG(PWRAP_LOG_ERROR,
 			  "Failed to copy %s - error: %s",
@@ -1203,55 +1306,55 @@ static int pwrap_pam_get_item(const pam_handle_t *pamh,
 		case PAM_USER:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item PAM_USER=%s",
-				  (char *) *item);
+				  (const char *)*item);
 			break;
 		case PAM_SERVICE:
 			svc = pwrap_get_service((const char *) *item);
 
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item PAM_SERVICE=%s",
-				  (char *) svc);
+				  svc);
 			*item = svc;
 			break;
 		case PAM_USER_PROMPT:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item PAM_USER_PROMPT=%s",
-				  (char *) *item);
+				  (const char *)*item);
 			break;
 		case PAM_TTY:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item PAM_TTY=%s",
-				  (char *) *item);
+				  (const char *)*item);
 			break;
 		case PAM_RUSER:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item PAM_RUSER=%s",
-				  (char *) *item);
+				  (const char *)*item);
 			break;
 		case PAM_RHOST:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item PAM_RHOST=%s",
-				  (char *) *item);
+				  (const char *)*item);
 			break;
 		case PAM_AUTHTOK:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item PAM_AUTHTOK=%s",
-				  (char *) *item);
+				  (const char *)*item);
 			break;
 		case PAM_OLDAUTHTOK:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item PAM_OLDAUTHTOK=%s",
-				  (char *) *item);
+				  (const char *)*item);
 			break;
 		case PAM_CONV:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item PAM_CONV=%p",
-				  (void *) *item);
+				  (const void *)*item);
 			break;
 		default:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_get_item item_type=%d item=%p",
-				  item_type, (void *) *item);
+				  item_type, (const void *)*item);
 			break;
 		}
 	} else {
@@ -1280,42 +1383,42 @@ static int pwrap_pam_set_item(pam_handle_t *pamh,
 		case PAM_USER:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_set_item PAM_USER=%s",
-				  (char *)item);
+				  (const char *)item);
 			break;
 		case PAM_SERVICE:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_set_item PAM_SERVICE=%s",
-				  (char *)item);
+				  (const char *)item);
 			break;
 		case PAM_USER_PROMPT:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_set_item PAM_USER_PROMPT=%s",
-				  (char *)item);
+				  (const char *)item);
 			break;
 		case PAM_TTY:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_set_item PAM_TTY=%s",
-				  (char *)item);
+				  (const char *)item);
 			break;
 		case PAM_RUSER:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_set_item PAM_RUSER=%s",
-				  (char *)item);
+				  (const char *)item);
 			break;
 		case PAM_RHOST:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_set_item PAM_RHOST=%s",
-				  (char *)item);
+				  (const char *)item);
 			break;
 		case PAM_AUTHTOK:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_set_item PAM_AUTHTOK=%s",
-				  (char *)item);
+				  (const char *)item);
 			break;
 		case PAM_OLDAUTHTOK:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
 				  "pwrap_set_item PAM_OLDAUTHTOK=%s",
-				  (char *)item);
+				  (const char *)item);
 			break;
 		case PAM_CONV:
 			PWRAP_LOG(PWRAP_LOG_TRACE,
@@ -1441,7 +1544,7 @@ int pam_prompt(pam_handle_t *pamh,
 			       args);
 	va_end(args);
 
-	return rv;  
+	return rv;
 }
 
 #ifdef HAVE_PAM_STRERROR_CONST
@@ -1474,16 +1577,58 @@ const char *pam_strerror(pam_handle_t *pamh, int errnum)
 				  errnum);
 }
 
-#ifdef HAVE_PAM_VSYSLOG
+#if defined(HAVE_PAM_VSYSLOG) || defined(HAVE_PAM_SYSLOG)
+static void pwrap_pam_vsyslog(const pam_handle_t *pamh,
+			      int priority,
+			      const char *fmt,
+			      va_list args) PRINTF_ATTRIBUTE(3, 0);
+
 static void pwrap_pam_vsyslog(const pam_handle_t *pamh,
 			      int priority,
 			      const char *fmt,
 			      va_list args)
 {
-	PWRAP_LOG(PWRAP_LOG_TRACE, "pwrap_pam_vsyslog called");
-	libpam_pam_vsyslog(pamh, priority, fmt, args);
-}
+	const char *d;
+	char syslog_str[32] = {0};
+	enum pwrap_dbglvl_e dbglvl = PWRAP_LOG_TRACE;
 
+	PWRAP_LOG(PWRAP_LOG_TRACE, "pwrap_pam_vsyslog called");
+
+#ifdef HAVE_PAM_VSYSLOG
+	d = getenv("PAM_WRAPPER_USE_SYSLOG");
+	if (d != NULL && d[0] == '1') {
+		libpam_pam_vsyslog(pamh, priority, fmt, args);
+		return;
+	}
+#endif /* HAVE_PAM_VSYSLOG */
+
+	switch(priority) {
+	case 0: /* LOG_EMERG */
+	case 1: /* LOG_ALERT */
+	case 2: /* LOG_CRIT */
+	case 3: /* LOG_ERR */
+		dbglvl = PWRAP_LOG_ERROR;
+		break;
+	case 4: /* LOG_WARN */
+		dbglvl = PWRAP_LOG_WARN;
+		break;
+	case 5: /* LOG_NOTICE */
+	case 6: /* LOG_INFO */
+	case 7: /* LOG_DEBUG */
+		dbglvl = PWRAP_LOG_DEBUG;
+		break;
+	default:
+		dbglvl = PWRAP_LOG_TRACE;
+		break;
+	}
+
+	snprintf(syslog_str, sizeof(syslog_str), "SYSLOG(%d)", priority);
+
+	pwrap_vlog(dbglvl, syslog_str, fmt, args);
+}
+#endif /* defined(HAVE_PAM_VSYSLOG) || defined(HAVE_PAM_SYSLOG) */
+
+#ifdef HAVE_PAM_VSYSLOG
 void pam_vsyslog(const pam_handle_t *pamh,
 		 int priority,
 		 const char *fmt,
